# This doesn't change much but occassionally
# check these values to ensure they reflect current prices.
PRICES = {
    # Price per 1 million characters based on the TTS engine used.
    'standard': 4.00,
    'neural': 16.00,
    'generative': 30.00,
    'long-form': 100.00,
}



import sys
from importlib.metadata import version

import typer

from jmestts.tts import (
    InvalidLanguageError,
    TextTooLongError,
    count_chars,
    create_tts_client,
    estimate_cost,
    list_polly_language_codes,
    resolve_tts_params,
    validate_max_chars,
)
from jmestts.voices import LANGUAGES

app = typer.Typer()


def _version_callback(value: bool) -> None:
    if value:
        typer.echo(f"jmes-tts {version('jmes-tts')}")
        raise typer.Exit()


def _print_languages() -> None:
    typer.echo('Language presets (--language):')
    for name in sorted(LANGUAGES):
        preset = LANGUAGES[name]
        typer.echo(
            f'- {name}: voice={preset["voice"]} engine={preset["engine"]} '
            f'language_code={preset["language_code"]}'
        )
    typer.echo()
    typer.echo('Polly language codes (--language-code):')
    for code in list_polly_language_codes():
        typer.echo(f'- {code}')


@app.command()
def tts(
    filename: str | None = typer.Option(
        None, help="Input file to convert to speech"
    ),
    text: str | None = typer.Option(None, help="Text to convert to speech"),
    bucket: str | None = typer.Option(
        None, help="S3 bucket for long-form text"
    ),
    output: str = typer.Option('output.mp3', help="Output audio file"),
    language: str | None = typer.Option(
        None, help="Language of phrase (e.g. english, french)"
    ),
    language_code: str | None = typer.Option(
        None, help="Language code override (e.g. en-US, fr-FR)"
    ),
    voice: str | None = typer.Option(
        None, help="Voice to use for text-to-speech"
    ),
    engine: str | None = typer.Option(
        None, help="TTS engine (neural or generative)"
    ),
    list_languages: bool = typer.Option(
        False,
        '--list-languages',
        help=('List language presets and Polly language codes, then exit'),
    ),
    dry_run: bool = typer.Option(
        False,
        help=(
            "Estimate cost and validate parameters without "
            "making any AWS calls"
        ),
    ),
    version: bool | None = typer.Option(
        None,
        "--version",
        callback=_version_callback,
        is_eager=True,
        help="Show version and exit",
    ),
) -> None:
    sys.excepthook = sys.__excepthook__

    if list_languages:
        _print_languages()
        raise typer.Exit(code=0)

    if text is None and filename is None:
        typer.echo(
            "Error: Either --filename or --text must be provided", err=True
        )
        raise typer.Exit(code=1)
    if text is not None and filename is not None:
        typer.echo(
            "Error: Provide either --filename or --text, but not both",
            err=True,
        )
        raise typer.Exit(code=1)

    contents: str
    if filename is not None:
        with open(filename) as f:
            contents = f.read()
    else:
        assert text is not None
        contents = text

    if dry_run:
        try:
            resolved = resolve_tts_params(
                language,
                voice=voice,
                engine=engine,
                language_code=language_code,
            )
            if bucket is None:
                validate_max_chars(contents)
        except (InvalidLanguageError, TextTooLongError) as exc:
            typer.echo(f"Error: {exc}", err=True)
            raise typer.Exit(code=1) from exc

        billable_chars = count_chars(contents)
        resolved_engine: str = resolved['engine']
        estimated_cost = estimate_cost(
            engine=resolved_engine, billable_chars=billable_chars
        )
        print(f"Num chars used: {billable_chars}")
        print(f"Total cost: ${estimated_cost:.6f} USD")
        print("Dry run: no AWS calls made.")
        raise typer.Exit(code=0)

    try:
        tts = create_tts_client(
            contents=text,
            filename=filename,
            bucket=bucket,
            language=language,
            language_code=language_code,
            voice=voice,
            engine=engine,
        )
    except InvalidLanguageError as exc:
        typer.echo(f"Error: {exc}", err=True)
        raise typer.Exit(code=1) from exc
    try:
        stream = tts.convert_to_speech(contents)
    except TextTooLongError as exc:
        typer.echo(f"Error: {exc}", err=True)
        raise typer.Exit(code=1) from exc
    print(f"Num chars used: {tts.last_request_chars}")
    print(f"Total cost: ${tts.last_cost:.6f} USD")
    with open(output, 'wb') as f:
        f.write(stream.read())
    print(f"Output written to {output}")


if __name__ == "__main__":
    app()



LANGUAGES = {
    'english': {
        'voice': 'Matthew',
        'engine': 'generative',
        'language_code': 'en-US',
    },
    'french': {
        'engine': 'generative',
        'voice': 'Lea',
        'language_code': 'fr-FR',
    },
    'spanish': {
        'engine': 'generative',
        'voice': 'Mia',
        'language_code': 'es-MX',
    },
    'cantonese': {
        'engine': 'neural',
        'voice': 'Hiujin',
        'language_code': 'yue-CN',
    },
    'mandarin': {
        'engine': 'neural',
        'voice': 'Zhiyu',
        'language_code': 'cmn-CN',
    },
}


def normalize_language(language: str) -> str:
    return language.strip().lower()



"""Performs text to speech using Amazon Polly."""

import re
import time
import logging
from collections.abc import Callable
from typing import Any, Optional, get_args

import boto3
from botocore.response import StreamingBody
from mypy_boto3_polly.literals import VoiceIdType, EngineType, LanguageCodeType
from mypy_boto3_polly.client import PollyClient
from mypy_boto3_s3.client import S3Client

from jmestts.voices import LANGUAGES, normalize_language
from jmestts.pricing import PRICES


LOG = logging.getLogger(__name__)

MAX_SYNC_BILLABLE_CHARS = 3000


def list_polly_language_codes() -> tuple[str, ...]:
    codes = (str(code) for code in get_args(LanguageCodeType))
    return tuple(sorted(set(codes)))


def count_chars(contents: str) -> int:
    """Count number of characters billed by Polly.

    Polly strips leading/trailing whitespace and collapses consecutive
    whitespace characters into a single space.
    """
    text = contents.strip()
    text = re.sub(r'\s+', ' ', text)
    return len(text)


class TextTooLongError(ValueError):
    pass


class InvalidLanguageError(ValueError):
    language: str
    valid_languages: tuple[str, ...]

    def __init__(
        self, language: str, *, valid_languages: tuple[str, ...]
    ) -> None:
        self.language = language
        self.valid_languages = valid_languages
        valid_languages_str = ', '.join(valid_languages)
        super().__init__(
            f"Invalid language: {language}. "
            f"Valid languages: {valid_languages_str}. "
            "Run --list-languages to see presets and Polly codes."
        )


def validate_max_chars(
    contents: str, *, max_chars: int = MAX_SYNC_BILLABLE_CHARS
) -> None:
    billable_chars = count_chars(contents)
    if billable_chars <= max_chars:
        return

    raise TextTooLongError(
        f"Text is {billable_chars} billable characters, which exceeds the "
        f"max of {max_chars} for synchronous TTS. Provide --bucket to use "
        "long-form synthesis, or shorten your input."
    )


def resolve_tts_params(
    language: Optional[str] = None,
    *,
    voice: Optional[str] = None,
    engine: Optional[str] = None,
    language_code: Optional[str] = None,
) -> dict[str, Any]:
    # Use 'english' as default if no language specified
    effective_language = language if language is not None else 'english'
    normalized_language = normalize_language(effective_language)
    if normalized_language not in LANGUAGES:
        valid_languages = tuple(sorted(LANGUAGES))
        raise InvalidLanguageError(
            effective_language, valid_languages=valid_languages
        )
    params = LANGUAGES[normalized_language]
    kwargs: dict[str, Any] = {**params}

    if voice is not None:
        kwargs['voice'] = voice
    if engine is not None:
        kwargs['engine'] = engine
    if language_code is not None:
        kwargs['language_code'] = language_code

    return kwargs


def estimate_cost(*, engine: str, billable_chars: int) -> float:
    if engine not in PRICES:
        return 0.0
    return PRICES[engine] * billable_chars / 1_000_000


def create_tts_client(
    contents: Optional[str] = None,
    filename: Optional[str] = None,
    bucket: Optional[str] = None,
    language: Optional[str] = None,
    language_code: Optional[str] = None,
    voice: Optional[str] = None,
    engine: Optional[str] = None,
) -> 'BaseTextToSpeech':
    if (contents is None and filename is None) or (
        contents is not None and filename is not None
    ):
        raise ValueError(
            "Exactly one of contents or filename must be provided"
        )
    kwargs = resolve_tts_params(
        language, voice=voice, engine=engine, language_code=language_code
    )

    if bucket is not None:
        cls = LongFormTextToSpeech
        kwargs['bucket'] = bucket
    else:
        cls = TextToSpeech
    return cls(**kwargs)


class BaseTextToSpeech:
    last_request_chars: int = 0
    engine: EngineType

    def convert_to_speech(self, contents: str) -> StreamingBody:
        raise NotImplementedError("convert_to_speech")

    @property
    def last_cost(self) -> float:
        if self.engine not in PRICES:
            return 0.0
        return PRICES[self.engine] * self.last_request_chars / 1_000_000


class TextToSpeech(BaseTextToSpeech):
    """Converts text to speech using Amazon Polly."""

    def __init__(
        self,
        polly_client: PollyClient | None = None,
        voice: VoiceIdType = 'Matthew',
        engine: EngineType = 'generative',
        language_code: LanguageCodeType = 'en-US',
    ) -> None:
        if polly_client is None:
            polly_client = boto3.client('polly')
        self._polly = polly_client
        self.voice: VoiceIdType = voice
        self.engine: EngineType = engine
        self.language_code: LanguageCodeType = language_code
        self.last_request_chars = 0

    def convert_to_speech(self, contents: str) -> StreamingBody:
        """Converts text to speech."""
        validate_max_chars(contents)
        response = self._polly.synthesize_speech(
            Text=contents,
            OutputFormat='mp3',
            VoiceId=self.voice,
            Engine=self.engine,
            LanguageCode=self.language_code,
        )
        self.last_request_chars = response['RequestCharacters']
        return response['AudioStream']


class LongFormTextToSpeech(BaseTextToSpeech):
    """Converts long form text to speech using Amazon Polly.

    When the text content is over a given size (currently 3000 chars for
    standard voice), you need to use this class to generate audio.

    This has a larger max character size from 100k to 200k, depending
    on the engine used.

    This class requires an async workflow where a job is started and then
    results are uploaded to an S3 bucket.  This class abstracts all those
    details to still provide a blocking API that will automatically
    download the results from S3 when complete, so you have a similar
    API to the sync version of this class.
    """

    DELAY = 5

    def __init__(
        self,
        bucket: str,
        polly_client: PollyClient | None = None,
        s3_client: S3Client | None = None,
        voice: VoiceIdType = 'Matthew',
        engine: EngineType = 'generative',
        language_code: LanguageCodeType = 'en-US',
        sleep: Callable[[float], None] | None = None,
    ) -> None:
        if polly_client is None:
            polly_client = boto3.client('polly')
        if s3_client is None:
            s3_client = boto3.client('s3')
        if sleep is None:
            sleep = time.sleep
        self._polly = polly_client
        self._s3 = s3_client
        self._sleep: Callable[[float], None] = sleep
        self.bucket = bucket
        self.voice: VoiceIdType = voice
        self.engine: EngineType = engine
        self.language_code: LanguageCodeType = language_code
        self.last_request_chars = 0

    def convert_to_speech(self, contents: str) -> StreamingBody:
        response = self._polly.start_speech_synthesis_task(
            Engine=self.engine,
            LanguageCode=self.language_code,
            OutputFormat='mp3',
            OutputS3BucketName=self.bucket,
            Text=contents,
            VoiceId=self.voice,
        )
        task_id = response['SynthesisTask']['TaskId']
        while True:
            # TODO: want to use waiters so we have timeouts/etc.
            result = self._polly.get_speech_synthesis_task(TaskId=task_id)
            status = result['SynthesisTask']['TaskStatus']
            if status == 'failed':
                raise RuntimeError(
                    f"TTS task failed, task_id={task_id}\n{result}"
                )
            elif status == 'completed':
                output_uri = result['SynthesisTask']['OutputUri']
                # The output uri is formatted as an https URL:
                # https://s3.us-east-1.amazonaws.com/jmes-tts/<pre>/<uuid>.mp3
                # So parse out the bucket and key.
                key = '/'.join(output_uri.split('/')[4:])
                stream = self._get_s3_download_stream(key)
                self.last_request_chars = result['SynthesisTask'][
                    'RequestCharacters'
                ]
                return stream
            self._sleep(self.DELAY)

    def _get_s3_download_stream(self, key: str):
        response = self._s3.get_object(Bucket=self.bucket, Key=key)
        return response['Body']



from pathlib import Path

from typer.testing import CliRunner

import jmestts.cli as cli
from jmestts.pricing import PRICES


def _deny_aws_calls(*_args: object, **_kwargs: object) -> object:
    raise AssertionError("dry-run must not create AWS clients")


def test_dry_run_prints_chars_and_estimated_cost(monkeypatch):
    runner = CliRunner()
    monkeypatch.setattr(cli, "create_tts_client", _deny_aws_calls)

    text = "Hello  World"
    result = runner.invoke(
        cli.app, ["--text", text, "--engine", "standard", "--dry-run"]
    )

    expected_chars = 11
    expected_cost = PRICES["standard"] * expected_chars / 1_000_000

    assert result.exit_code == 0, result.output
    assert f"Num chars used: {expected_chars}" in result.output
    assert f"Total cost: ${expected_cost:.6f} USD" in result.output
    assert "Dry run: no AWS calls made." in result.output


def test_dry_run_does_not_write_output_file(monkeypatch):
    runner = CliRunner()
    monkeypatch.setattr(cli, "create_tts_client", _deny_aws_calls)

    with runner.isolated_filesystem():
        output_path = Path("dry_run.mp3")
        result = runner.invoke(
            cli.app,
            [
                "--text",
                "Hello",
                "--output",
                str(output_path),
                "--dry-run",
            ],
        )

        assert result.exit_code == 0, result.output
        assert not output_path.exists()


def test_dry_run_validates_3k_limit_for_sync_cases(monkeypatch):
    runner = CliRunner()
    monkeypatch.setattr(cli, "create_tts_client", _deny_aws_calls)

    result = runner.invoke(cli.app, ["--text", "A" * 3001, "--dry-run"])

    assert result.exit_code == 1
    assert "exceeds the max of 3000" in result.output
    assert "--bucket" in result.output


def test_dry_run_skips_3k_validation_when_bucket_provided(monkeypatch):
    runner = CliRunner()
    monkeypatch.setattr(cli, "create_tts_client", _deny_aws_calls)

    result = runner.invoke(
        cli.app, ["--text", "A" * 3001, "--bucket", "b", "--dry-run"]
    )

    assert result.exit_code == 0, result.output
    assert "Num chars used: 3001" in result.output



from typer.testing import CliRunner

import jmestts.cli as cli
from jmestts.voices import LANGUAGES


def test_unknown_language_lists_valid_languages() -> None:
    runner = CliRunner()
    result = runner.invoke(
        cli.app,
        [
            '--text',
            'Hello',
            '--language',
            'klingon',
            '--dry-run',
        ],
    )

    assert result.exit_code == 1
    assert "Error: Invalid language: klingon." in result.output
    assert "Valid languages:" in result.output
    assert "Traceback" not in result.output
    for language in sorted(LANGUAGES):
        assert language in result.output



from collections.abc import Callable
from io import BytesIO

import boto3
import pytest
from botocore.stub import Stubber

from jmestts import tts


@pytest.fixture(autouse=True)
def mock_aws_credentials(monkeypatch):
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "testing")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "testing")
    monkeypatch.setenv("AWS_SESSION_TOKEN", "testing")
    monkeypatch.setenv("AWS_DEFAULT_REGION", "us-east-1")
    monkeypatch.setenv("AWS_REGION", "us-east-1")


@pytest.fixture
def noop_sleep() -> Callable[[float], None]:
    def noop(_: float) -> None:
        return None

    return noop


@pytest.fixture
def polly_client_stub():
    polly_client = boto3.client('polly', region_name='us-east-1')
    stubber = Stubber(polly_client)
    stubber.activate()
    yield polly_client, stubber
    stubber.deactivate()


@pytest.fixture
def s3_client_stub():
    s3_client = boto3.client('s3', region_name='us-east-1')
    stubber = Stubber(s3_client)
    stubber.activate()
    yield s3_client, stubber
    stubber.deactivate()


# Test for count_chars function


@pytest.mark.parametrize(
    "input_text, expected_count",
    [
        ("Hello World", 11),
        ("", 0),
        ("A" * 100, 100),
        ("\n\t ", 0),
        # Consecutive spaces collapsed to single space
        ("Hello  World", 11),
        ("Hello    World", 11),
        # Tabs collapsed to single space
        ("Hello\tWorld", 11),
        ("Hello\t\tWorld", 11),
        # Newlines collapsed to single space
        ("Hello\nWorld", 11),
        ("Hello\n\nWorld", 11),
        # Mixed whitespace collapsed to single space
        ("Hello \t\n World", 11),
        ("One  Two   Three", 13),
        # Leading/trailing whitespace with internal collapse
        ("  Hello  World  ", 11),
        ("\n\tHello\t\tWorld\n\t", 11),
    ],
)
def test_count_chars(input_text, expected_count):
    assert tts.count_chars(input_text) == expected_count


def test_validate_max_chars_uses_billable_chars():
    contents = "A" + (" " * 4000) + "B"
    tts.validate_max_chars(contents)


def test_validate_max_chars_over_limit_raises_helpful_error():
    contents = "A" * (tts.MAX_SYNC_BILLABLE_CHARS + 1)
    with pytest.raises(
        tts.TextTooLongError,
        match=r"Text is 3001 billable characters.*max of 3000.*--bucket",
    ):
        tts.validate_max_chars(contents)


def test_tts_convert_to_speech_over_limit_errors(polly_client_stub):
    polly_client, _ = polly_client_stub
    tts_client = tts.TextToSpeech(polly_client=polly_client)
    contents = "A" * (tts.MAX_SYNC_BILLABLE_CHARS + 1)
    with pytest.raises(tts.TextTooLongError, match=r"max of 3000"):
        tts_client.convert_to_speech(contents)


def test_resolve_tts_params_overrides_voice_and_engine():
    resolved = tts.resolve_tts_params(
        " English ", voice="Joanna", engine="standard"
    )
    assert resolved["voice"] == "Joanna"
    assert resolved["engine"] == "standard"
    assert resolved["language_code"] == "en-US"


def test_estimate_cost_unknown_engine_is_zero():
    assert (
        tts.estimate_cost(engine="unknown_engine", billable_chars=1000) == 0.0
    )


def test_tts_convert_to_speech(polly_client_stub):
    polly_client, stubber = polly_client_stub
    test_text = "Hello, this is a test."
    expected_params = {
        'Text': test_text,
        'OutputFormat': 'mp3',
        'VoiceId': 'Matthew',
        'Engine': 'generative',
        'LanguageCode': 'en-US',
    }
    mock_stream = BytesIO(b"fake audio data")
    response = {
        'RequestCharacters': len(test_text),
        'AudioStream': mock_stream,
    }
    stubber.add_response('synthesize_speech', response, expected_params)

    tts_client = tts.TextToSpeech(polly_client=polly_client)
    stream = tts_client.convert_to_speech(test_text)

    assert stream.read() == b"fake audio data"
    assert tts_client.last_request_chars == len(test_text)


def test_long_form_tts_convert_to_speech(
    polly_client_stub, s3_client_stub, noop_sleep
):
    polly_client, polly_stubber = polly_client_stub
    s3_client, s3_stubber = s3_client_stub
    bucket_name = "test-bucket"
    test_text = (
        "This is a long form text." * 200
    )  # Ensures over 3000 chars if needed

    # Mock response for start_speech_synthesis_task
    start_response = {
        'SynthesisTask': {
            'TaskId': 'test-task-id',
            'TaskStatus': 'inProgress',
            'OutputUri': (
                f"https://s3.us-east-1.amazonaws.com/{bucket_name}/"
                "path/to/output.mp3"
            ),
            'RequestCharacters': len(test_text),
        }
    }
    start_params = {
        'Engine': 'generative',
        'LanguageCode': 'en-US',
        'OutputFormat': 'mp3',
        'OutputS3BucketName': bucket_name,
        'Text': test_text,
        'VoiceId': 'Matthew',
    }
    polly_stubber.add_response(
        'start_speech_synthesis_task', start_response, start_params
    )

    # Mock response for get_speech_synthesis_task (first call: inProgress)
    get_response_in_progress = {
        'SynthesisTask': {
            'TaskId': 'test-task-id',
            'TaskStatus': 'inProgress',
            'OutputUri': (
                f"https://s3.us-east-1.amazonaws.com/{bucket_name}"
                "/path/to/output.mp3"
            ),
            'RequestCharacters': len(test_text),
        }
    }
    # Mock response for get_speech_synthesis_task (second call: completed)
    get_response_completed = {
        'SynthesisTask': {
            'TaskId': 'test-task-id',
            'TaskStatus': 'completed',
            'OutputUri': (
                f"https://s3.us-east-1.amazonaws.com/{bucket_name}"
                "/path/to/output.mp3"
            ),
            'RequestCharacters': len(test_text),
        }
    }
    get_params = {
        'TaskId': 'test-task-id',
    }
    polly_stubber.add_response(
        'get_speech_synthesis_task', get_response_in_progress, get_params
    )
    polly_stubber.add_response(
        'get_speech_synthesis_task', get_response_completed, get_params
    )

    # Mock response for s3.get_object
    s3_key = "path/to/output.mp3"
    s3_response = {
        'Body': BytesIO(b"fake long form audio data"),
    }
    s3_stubber.add_response(
        'get_object',
        s3_response,
        {
            'Bucket': bucket_name,
            'Key': s3_key,
        },
    )

    # Initialize the testable LongFormTextToSpeech with stubbed clients
    tts_client = tts.LongFormTextToSpeech(
        bucket=bucket_name,
        polly_client=polly_client,
        s3_client=s3_client,
        sleep=noop_sleep,
    )

    stream = tts_client.convert_to_speech(test_text)

    assert stream.read() == b"fake long form audio data"
    assert tts_client.last_request_chars == len(test_text)


def test_long_form_tts_task_failure(
    polly_client_stub, s3_client_stub, noop_sleep
):
    polly_client, polly_stubber = polly_client_stub
    s3_client, _ = s3_client_stub
    bucket_name = "test-bucket"
    test_text = "This is a long form text." * 200
    start_response = {
        'SynthesisTask': {
            'TaskId': 'test-task-id',
            'TaskStatus': 'inProgress',
            'OutputUri': (
                f"https://s3.us-east-1.amazonaws.com/{bucket_name}"
                "/path/to/output.mp3"
            ),
            'RequestCharacters': len(test_text),
        }
    }
    start_params = {
        'Engine': 'generative',
        'LanguageCode': 'en-US',
        'OutputFormat': 'mp3',
        'OutputS3BucketName': bucket_name,
        'Text': test_text,
        'VoiceId': 'Matthew',
    }
    polly_stubber.add_response(
        'start_speech_synthesis_task', start_response, start_params
    )

    # Mock response for get_speech_synthesis_task indicating failure
    get_response_failed = {
        'SynthesisTask': {
            'TaskId': 'test-task-id',
            'TaskStatus': 'failed',
            'OutputUri': (
                f"https://s3.us-east-1.amazonaws.com/{bucket_name}"
                "/path/to/output.mp3"
            ),
            'RequestCharacters': len(test_text),
        }
    }
    get_params = {
        'TaskId': 'test-task-id',
    }
    polly_stubber.add_response(
        'get_speech_synthesis_task', get_response_failed, get_params
    )

    # Initialize the testable LongFormTextToSpeech with stubbed clients
    tts_client = tts.LongFormTextToSpeech(
        bucket=bucket_name,
        polly_client=polly_client,
        s3_client=s3_client,
        sleep=noop_sleep,
    )

    with pytest.raises(RuntimeError, match="TTS task failed"):
        tts_client.convert_to_speech(test_text)


# Test for create_tts_client function


def test_create_tts_client_with_contents_only():
    client = tts.create_tts_client(contents="Hello world")
    assert isinstance(client, tts.TextToSpeech)
    assert client.voice == 'Matthew'
    assert client.engine == 'generative'
    assert client.language_code == 'en-US'


def test_create_tts_client_with_bucket():
    client = tts.create_tts_client(
        contents="Hello world", bucket="test-bucket"
    )
    assert isinstance(client, tts.LongFormTextToSpeech)
    assert client.bucket == "test-bucket"
    assert client.voice == 'Matthew'
    assert client.engine == 'generative'
    assert client.language_code == 'en-US'


def test_create_tts_client_with_different_language():
    client = tts.create_tts_client(contents="Bonjour", language='french')
    assert isinstance(client, tts.TextToSpeech)
    assert client.voice == 'Lea'
    assert client.engine == 'generative'
    assert client.language_code == 'fr-FR'


def test_create_tts_client_invalid_language():
    with pytest.raises(ValueError, match="Invalid language: invalid_lang"):
        tts.create_tts_client(contents="Hello", language='invalid_lang')


def test_create_tts_client_no_contents_or_filename():
    with pytest.raises(
        ValueError,
        match="Exactly one of contents or filename must be provided",
    ):
        tts.create_tts_client(language='english')


def test_last_cost_calculation(polly_client_stub):
    polly_client, stubber = polly_client_stub
    test_text = "Hello" * 100  # 500 chars

    expected_params = {
        'Text': test_text,
        'OutputFormat': 'mp3',
        'VoiceId': 'Matthew',
        'Engine': 'generative',
        'LanguageCode': 'en-US',
    }
    mock_stream = BytesIO(b"fake audio data")
    response = {
        'RequestCharacters': 500,
        'AudioStream': mock_stream,
    }
    stubber.add_response('synthesize_speech', response, expected_params)

    tts_client = tts.TextToSpeech(polly_client=polly_client)
    tts_client.convert_to_speech(test_text)

    expected_cost = 30 * 500 / 1_000_000
    assert tts_client.last_cost == expected_cost


def test_last_cost_unknown_engine():
    tts_client = tts.TextToSpeech()
    tts_client.engine = 'unknown_engine'
    tts_client.last_request_chars = 1000
    assert tts_client.last_cost == 0.0


def test_text_to_speech_custom_params(polly_client_stub):
    polly_client, stubber = polly_client_stub
    test_text = "Hello world"

    expected_params = {
        'Text': test_text,
        'OutputFormat': 'mp3',
        'VoiceId': 'Joanna',
        'Engine': 'long-form',
        'LanguageCode': 'en-GB',
    }
    mock_stream = BytesIO(b"fake audio data")
    response = {
        'RequestCharacters': len(test_text),
        'AudioStream': mock_stream,
    }
    stubber.add_response('synthesize_speech', response, expected_params)

    tts_client = tts.TextToSpeech(
        polly_client=polly_client,
        voice='Joanna',
        engine='long-form',
        language_code='en-GB',
    )
    stream = tts_client.convert_to_speech(test_text)

    assert stream.read() == b"fake audio data"
    assert tts_client.last_request_chars == len(test_text)



from typer.testing import CliRunner

import jmestts.cli as cli
from jmestts.tts import list_polly_language_codes
from jmestts.voices import LANGUAGES


def test_list_languages_prints_presets_and_polly_codes() -> None:
    runner = CliRunner()
    result = runner.invoke(cli.app, ['--list-languages'])

    assert result.exit_code == 0, result.output

    output_lines = result.output.splitlines()
    assert 'Language presets (--language):' in output_lines
    for name in LANGUAGES:
        assert any(line.startswith(f'- {name}:') for line in output_lines)
    assert 'Polly language codes (--language-code):' in output_lines
    for code in list_polly_language_codes():
        assert f'- {code}' in output_lines
